
from wofry.propagator.propagators2D.fraunhofer import Fraunhofer2D
from wofry.propagator.propagators2D.fresnel import Fresnel2D
from wofry.propagator.propagators2D.fresnel_convolution import FresnelConvolution2D
from wofry.propagator.propagators2D.integral import Integral2D
from wofry.propagator.propagators2D.fresnel_zoom_xy import FresnelZoomXY2D
from wofry.propagator.propagator import PropagationManager

def initialize_default_propagator_2D():
    propagator = PropagationManager.Instance()

    propagator.add_propagator(Fraunhofer2D())
    propagator.add_propagator(Fresnel2D())
    propagator.add_propagator(FresnelConvolution2D())
    propagator.add_propagator(Integral2D())
    propagator.add_propagator(FresnelZoomXY2D())



# TODO: REMOVE THIS!!!!
try:
    from oasys_srw.srwlib import *
    SRWLIB_AVAILABLE = True
except:
    try:
        from srwlib import *
        SRWLIB_AVAILABLE = True
    except:
        SRWLIB_AVAILABLE = False
        print("SRW is not available")

import scipy.constants as codata
angstroms_to_eV = codata.h*codata.c/codata.e*1e10

from wofry.propagator.wavefront2D.generic_wavefront import GenericWavefront2D
from wofry.propagator.propagator import Propagator2D
from wofry.propagator.examples import WOSRWWavefront

class FresnelSRW(Propagator2D):

    HANDLER_NAME = "FRESNEL_SRW"

    def get_handler_name(self):
        return self.HANDLER_NAME

    """
    2D Fresnel propagator using convolution via Fourier transform
    :param wavefront:
    :param propagation_distance:
    :param srw_autosetting:set to 1 for automatic SRW redimensionate wavefront
    :return:
    """

    def do_specific_progation(self, wavefront, propagation_distance, parameters):
        if not SRWLIB_AVAILABLE: raise ImportError("SRW is not available")

        if not parameters.has_additional_parameter("srw_autosetting"):
            srw_autosetting = 0
        else:
            srw_autosetting = parameters.get_additional_parameter("srw_autosetting")

        is_generic_wavefront = isinstance(wavefront, GenericWavefront2D)

        if is_generic_wavefront:
            wavefront = WOSRWWavefront.fromGenericWavefront(wavefront)
        else:
            if not isinstance(wavefront, WOSRWWavefront): raise ValueError("wavefront cannot be managed by this propagator")

        #
        # propagation
        #
        optDrift = SRWLOptD(propagation_distance) #Drift space


        #Wavefront Propagation Parameters:
        #[0]: Auto-Resize (1) or not (0) Before propagation
        #[1]: Auto-Resize (1) or not (0) After propagation
        #[2]: Relative Precision for propagation with Auto-Resizing (1. is nominal)
        #[3]: Allow (1) or not (0) for semi-analytical treatment of the quadratic (leading) phase terms at the propagation
        #[4]: Do any Resizing on Fourier side, using FFT, (1) or not (0)
        #[5]: Horizontal Range modification factor at Resizing (1. means no modification)
        #[6]: Horizontal Resolution modification factor at Resizing
        #[7]: Vertical Range modification factor at Resizing
        #[8]: Vertical Resolution modification factor at Resizing
        #[9]: Type of wavefront Shift before Resizing (not yet implemented)
        #[10]: New Horizontal wavefront Center position after Shift (not yet implemented)
        #[11]: New Vertical wavefront Center position after Shift (not yet implemented)

        if srw_autosetting:
            #                 0  1  2   3  4  5   6   7   8   9 10 11
            propagParDrift = [1, 1, 1., 0, 0, 1., 1., 1., 1., 0, 0, 0]
        else:
            #                 0  1  2   3  4  5   6   7   8   9 10 11
            propagParDrift = [0, 0, 1., 0, 0, 1., 1., 1., 1., 0, 0, 0]

        optBL = SRWLOptC([optDrift], [propagParDrift]) #"Beamline" - Container of Optical Elements (together with the corresponding wavefront propagation instructions)

        print('   Simulating Electric Field Wavefront Propagation by SRW ... ', end='\n')
        srwl.PropagElecField(wavefront, optBL)

        if is_generic_wavefront:
            return wavefront.toGenericWavefront()
        else:
            return wavefront



import numpy
import scipy.constants as codata

from srxraylib.util.data_structures import ScaledMatrix, ScaledArray

from wofry.propagator.wavefront import Wavefront, WavefrontDimension
from wofry.propagator.wavefront1D.generic_wavefront import GenericWavefront1D

from wofry.propagator.util.gaussian_schell_model import GaussianSchellModel2D

import copy


from wofry.propagator.polarization import Polarization
# class Polarization:
#     SIGMA = 0
#     PI = 1
#     TOTAL = 3

# needed for h5 i/o
import os
import sys
import time
try:
    import h5py
except:
    raise ImportError("h5py not available: input/output to files not working")
# --------------------------------------------------
# Wavefront 2D
# --------------------------------------------------




class GenericWavefront2D(Wavefront):
    XX = 0
    YY = 1


    def __init__(self, wavelength=1e-10, electric_field_matrix=None, electric_field_matrix_pi=None):
        self._wavelength = wavelength
        self._electric_field_matrix = electric_field_matrix
        self._electric_field_matrix_pi = electric_field_matrix_pi

    def get_dimension(self):
        return WavefrontDimension.TWO

    def is_polarized(self):
        if self._electric_field_matrix_pi is None:
            return False
        else:
            return True

    def duplicate(self):

        if self.is_polarized():
            return GenericWavefront2D(wavelength=self._wavelength,
                                      electric_field_matrix=ScaledMatrix(x_coord=   copy.copy(self._electric_field_matrix.x_coord),
                                                                         y_coord=   copy.copy(self._electric_field_matrix.y_coord),
                                                                         z_values=  copy.copy(self._electric_field_matrix.z_values),
                                                                         interpolator=        self._electric_field_matrix.interpolator),
                                      electric_field_matrix_pi=ScaledMatrix(x_coord=copy.copy(self._electric_field_matrix_pi.x_coord),
                                                                         y_coord=   copy.copy(self._electric_field_matrix_pi.y_coord),
                                                                         z_values=  copy.copy(self._electric_field_matrix_pi.z_values),
                                                                         interpolator=        self._electric_field_matrix_pi.interpolator),
                                      )
        else:
            return GenericWavefront2D(wavelength=self._wavelength,
                                      electric_field_matrix=ScaledMatrix(x_coord=copy.copy(self._electric_field_matrix.x_coord),
                                                                         y_coord=copy.copy(self._electric_field_matrix.y_coord),
                                                                         z_values=copy.copy(self._electric_field_matrix.z_values),
                                                                         interpolator=self._electric_field_matrix.interpolator))

    @classmethod
    def initialize_wavefront(cls, number_of_points=(100,100), wavelength=1e-10, polarization=Polarization.SIGMA):

        sM = ScaledMatrix.initialize(np_array_z=numpy.full(number_of_points, (1.0 + 0.0j),dtype=complex),
                                                          interpolator=False)

        if ((polarization == Polarization.PI) or (polarization == Polarization.TOTAL)):
            sM_pi = ScaledMatrix.initialize(np_array_z=numpy.full(number_of_points, (0.0 + 0.0j),dtype=complex),
                                                          interpolator=False)
        else:
            sM_pi = None

        return GenericWavefront2D(wavelength, sM, sM_pi)

    @classmethod
    def initialize_wavefront_from_steps(cls, x_start=0.0, x_step=1.0, y_start=0.0, y_step=1.0,
                                        number_of_points=(100,100), wavelength=1e-10,
                                        polarization=Polarization.SIGMA):
        sM = ScaledMatrix.initialize_from_steps(numpy.full(number_of_points,(1.0 + 0.0j),dtype=complex),
                                                x_start,
                                                x_step,
                                                y_start,
                                                y_step,
                                                interpolator=False)

        if ((polarization == Polarization.PI) or (polarization == Polarization.TOTAL)):
            sM_pi = ScaledMatrix.initialize_from_steps(numpy.full(number_of_points,(0.0 + 0.0j),dtype=complex),
                                                x_start,
                                                x_step,
                                                y_start,
                                                y_step,
                                                interpolator=False)
        else:
            sM_pi = None

        return GenericWavefront2D(wavelength,sM, sM_pi)

    @classmethod
    def initialize_wavefront_from_range(cls, x_min=0.0, x_max=0.0, y_min=0.0, y_max=0.0,
                                        number_of_points=(100,100), wavelength=1e-10,
                                        polarization=Polarization.SIGMA ):

        sM = ScaledMatrix.initialize_from_range( \
                    numpy.full(number_of_points, (1.0 + 0.0j), dtype=complex),
                    x_min,x_max,y_min,y_max,interpolator=False)

        if ((polarization == Polarization.PI) or (polarization == Polarization.TOTAL)):
            sM_pi = ScaledMatrix.initialize_from_range( \
                    numpy.full(number_of_points, (0.0 + 0.0j), dtype=complex),
                    x_min,x_max,y_min,y_max,interpolator=False)
        else:
            sM_pi = None

        return GenericWavefront2D(wavelength, sM, sM_pi)

    @classmethod
    def initialize_wavefront_from_arrays(cls,x_array, y_array,  z_array, z_array_pi=None, wavelength=1e-10):
        sh = z_array.shape

        if sh[0] != x_array.size:
            raise Exception("Unmatched shapes for x")
        
        if sh[1] != y_array.size:
            raise Exception("Unmatched shapes for y")

        sM = ScaledMatrix.initialize_from_steps(
                    z_array,x_array[0],numpy.abs(x_array[1]-x_array[0]),
                            y_array[0],numpy.abs(y_array[1]-y_array[0]),interpolator=False)

        if z_array_pi is None:
            sM_pi = None
        else:
            sh = z_array_pi.shape
            if sh[0] != x_array.size:
                raise Exception("Unmatched shapes for x (Pi Polarization)")

            if sh[1] != y_array.size:
                raise Exception("Unmatched shapes for y (Pi Polarization)")

            sM_pi = ScaledMatrix.initialize_from_steps(
                        z_array_pi,x_array[0],numpy.abs(x_array[1]-x_array[0]),
                                y_array[0],numpy.abs(y_array[1]-y_array[0]),interpolator=False)

        return GenericWavefront2D(wavelength, sM, sM_pi)



    # TODO: polarization
    def get_Wavefront1D_from_profile(self, axis, coordinate):
        # swap axis - changed giovanni+manuel
        if axis == 1: # fixed X
            index = numpy.argmin(numpy.abs(self._electric_field_matrix.x_coord - coordinate))

            return GenericWavefront1D(wavelength=self._wavelength,
                                      electric_field_array=ScaledArray(scale=self._electric_field_matrix.y_coord,
                                                                       np_array=self._electric_field_matrix.z_values[index, :]))
        elif axis == 0:
            index = numpy.argmin(numpy.abs(self._electric_field_matrix.y_coord - coordinate))

            return GenericWavefront1D(wavelength=self._wavelength,
                                      electric_field_array=ScaledArray(scale=self._electric_field_matrix.x_coord,
                                                                       np_array=self._electric_field_matrix.z_values[:, index]))

    #TODO
    def get_Wavefront1D_from_histogram(self, axis):
        raise NotImplementedError("Not yet implemented!")


    # TODO: polarization
    @classmethod
    def combine_1D_wavefronts_into_2D(cls, wavefront_h, wavefront_v, normalize_to=0, wavelength=0.0):
        if not wavelength > 0.0:
            wavelength = (wavefront_h.get_wavelength() + wavefront_v.get_wavelength())/2

        if normalize_to == 0:
            normalization_factor = numpy.sqrt(numpy.sum(wavefront_h.get_intensity()))
        elif normalize_to == 1:
            normalization_factor = numpy.sqrt(numpy.sum(wavefront_v.get_intensity()))
        else:
            normalization_factor = 1.0

        wavefront_2D = GenericWavefront2D.initialize_wavefront_from_steps(x_start=wavefront_h.offset(),
                                                                          x_step=wavefront_h.delta(),
                                                                          y_start=wavefront_v.offset(),
                                                                          y_step=wavefront_v.delta(),
                                                                          number_of_points=(wavefront_h.size(), wavefront_v.size()))

        # complex_amplitude =  numpy.zeros((wavefront_h.size(), wavefront_v.size()), dtype=complex)
        #
        # for i in range (0, wavefront_h.size()):
        #     for j in range (0, wavefront_v.size()):
        #         complex_amplitude[i, j] = complex(wavefront_h.get_amplitude()[i]*wavefront_v.get_amplitude()[j],
        #                                           wavefront_h.get_phase()[i] + wavefront_v.get_phase()[j])

        complex_amplitude = numpy.outer(wavefront_h.get_complex_amplitude(), wavefront_v.get_complex_amplitude())

        normalization_factor /= numpy.sum(numpy.abs(complex_amplitude))

        wavefront_2D.set_complex_amplitude(complex_amplitude * normalization_factor)

        return wavefront_2D

    #
    # main parameters
    #

    # grid

    def size(self):
        return self._electric_field_matrix.shape()

    def delta(self):
        x = self.get_coordinate_x()
        y = self.get_coordinate_y()
        return numpy.abs(x[1]-x[0]),numpy.abs(y[1]-y[0])
    #
    def offset(self):
        return self.get_coordinate_x()[0],self.get_coordinate_y()[0]


    def get_coordinate_x(self):
        return self._electric_field_matrix.get_x_values()

    def get_coordinate_y(self):
        return self._electric_field_matrix.get_y_values()


    def get_mesh_indices_x(self):
        return numpy.outer( numpy.arange(0,self.size()[0]), numpy.ones(self.size()[1]))

    def get_mesh_indices_y(self):
        return numpy.outer( numpy.ones(self.size()[0]), numpy.arange(0,self.size()[1]))

    def get_mask_grid(self,width_in_pixels=(1,1),number_of_lines=(1,1)):
        """

        :param width_in_pixels: (pixels_for_horizontal_lines,pixels_for_vertical_lines
        :param number_of_lines: (number_of_horizontal_lines, number_of_vertical_lines)
        :return:
        """

        indices_x = self.get_mesh_indices_x()
        indices_y = self.get_mesh_indices_y()

        used_indices = numpy.zeros( self.size(),dtype=int)

        for i in range(number_of_lines[1]):
            used_indices[ numpy.where( numpy.abs(indices_x - (i+1)*self.size()[0]/(1+number_of_lines[1])) <= (width_in_pixels[1]-1) )] = 1
        for i in range(number_of_lines[0]):
            used_indices[ numpy.where( numpy.abs(indices_y - (i+1)*self.size()[1]/(1+number_of_lines[0])) <= (width_in_pixels[0]-1) )] = 1

        return used_indices

    # only for 2D
    def get_mesh_x(self):
        XY = numpy.meshgrid(self.get_coordinate_x(),self.get_coordinate_y())
        return XY[0].T

    def get_mesh_y(self):
        XY = numpy.meshgrid(self.get_coordinate_x(),self.get_coordinate_y())
        return XY[1].T

    # photon energy

    def get_wavelength(self):
        return self._wavelength

    def get_wavenumber(self):
        return 2*numpy.pi/self._wavelength

    def get_photon_energy(self):
        m2ev = codata.c * codata.h / codata.e      # lambda(m)  = m2eV / energy(eV)
        return  m2ev / self._wavelength


    # wavefront

    def get_complex_amplitude(self, polarization=Polarization.SIGMA):

        if polarization == Polarization.SIGMA:
            return self._electric_field_matrix.get_z_values()
        elif polarization == Polarization.PI:
            if self.is_polarized():
                return self._electric_field_matrix_pi.get_z_values()
            else:
                raise Exception("Wavefront is not polarized.")
        else:
            raise Exception("Only 0=SIGMA and 1=PI are valid polarization values.")


    def get_amplitude(self, polarization=Polarization.SIGMA):
        return numpy.absolute(self.get_complex_amplitude(polarization=polarization))

    def get_phase(self,from_minimum_intensity=0.0,unwrap=0, polarization=Polarization.SIGMA):

        """

        :param from_minimum_intensity: set to zero phase values at pixels where intensity
                                        is less than from_minimum_intensity threshold
        :param unwrap: Flag to unwrap the returned phase:
            0: No unwrap (default)
            1: Unwrap only in Horizontal axis.
            2: Unwrap only in Vertical axis.
            3: Unwrap first in H, then in V.
            4: Unwrap first in V, then in H.
        :param polarization: 0=Sigma, 1=Pi
        :return: the phase in a numpy array
        """
        # return numpy.arctan2(numpy.imag(self.get_complex_amplitude()), numpy.real(self.get_complex_amplitude()))
        phase = numpy.angle( self.get_complex_amplitude(polarization=polarization) )

        if (from_minimum_intensity > 0.0):
            intensity = self.get_intensity()
            intensity /= intensity.max()
            bad_indices = numpy.where(intensity < from_minimum_intensity )
            phase[bad_indices] = 0.0

        if unwrap > 0:
            if unwrap == 1: # x only
                phase = numpy.unwrap(phase,axis=0)
            elif unwrap == 2: # y only
                phase = numpy.unwrap(phase,axis=1)
            elif unwrap == 3: # x and y
                phase = numpy.unwrap(numpy.unwrap(phase,axis=0),axis=1)
            elif unwrap == 4: # y and x
                phase = numpy.unwrap(numpy.unwrap(phase,axis=1),axis=0)
            else:
                raise Exception(NotImplemented)

        return phase

    def get_intensity(self, polarization=Polarization.SIGMA):

        if polarization == Polarization.TOTAL:
            if self.is_polarized():
                return self.get_amplitude(polarization=Polarization.SIGMA)**2 + \
                       self.get_amplitude(polarization=Polarization.PI)**2
            else:
                return self.get_amplitude(polarization=Polarization.SIGMA)**2
        else:
            return self.get_amplitude(polarization=polarization)**2


    def get_integrated_intensity(self, polarization=Polarization.SIGMA):
        deltas =    (self.get_coordinate_x()[1] - self.get_coordinate_x()[0]) * \
                    (self.get_coordinate_y()[1] - self.get_coordinate_y()[0])
        return self.get_intensity(polarization=polarization).sum() * deltas



    # interpolated values

    def get_interpolated_complex_amplitude(self, x_value,y_value, polarization=Polarization.SIGMA):

        if polarization == Polarization.SIGMA:
            return self._electric_field_matrix.interpolate_value(x_value,y_value)
        elif polarization == Polarization.PI:
            if self.is_polarized():
                return self._electric_field_matrix_pi.interpolate_value(x_value,y_value)
            else:
                raise Exception("Wavefront is not polarized.")
        else:
            raise Exception("Only 0=SIGMA and 1=PI are valid polarization values.")

    def get_interpolated_amplitude(self, x_value,y_value, polarization=Polarization.SIGMA): # singular!
        interpolated_complex_amplitude = self.get_interpolated_complex_amplitude(x_value,y_value,polarization=polarization)
        return numpy.abs(interpolated_complex_amplitude)

    def get_interpolated_phase(self, x_value,y_value, polarization=Polarization.SIGMA): # singular!
        interpolated_complex_amplitude = self.get_interpolated_complex_amplitude(x_value,y_value,polarization=polarization)
        return numpy.arctan2(numpy.imag(interpolated_complex_amplitude), numpy.real(interpolated_complex_amplitude))


    def get_interpolated_intensity(self, x_value,y_value, polarization=Polarization.SIGMA):
        if polarization == Polarization.TOTAL:
            interpolated_complex_amplitude = self.get_interpolated_complex_amplitude(x_value,y_value,polarization=Polarization.SIGMA)
            if self.is_polarized():
                interpolated_complex_amplitude_pi = self.get_interpolated_complex_amplitude(x_value,y_value,polarization=Polarization.PI)
                return numpy.abs(interpolated_complex_amplitude)**2 + numpy.abs(interpolated_complex_amplitude_pi)**2
            else:
                return numpy.abs(interpolated_complex_amplitude)**2
        elif polarization == Polarization.SIGMA:
            interpolated_complex_amplitude = self.get_interpolated_complex_amplitude(x_value,y_value,polarization=Polarization.SIGMA)
            return numpy.abs(interpolated_complex_amplitude)**2
        elif polarization == Polarization.PI:
            interpolated_complex_amplitude_pi = self.get_interpolated_complex_amplitude(x_value,y_value,polarization=Polarization.PI)
            return numpy.abs(interpolated_complex_amplitude_pi)**2
        else:
            raise Exception("Wrong polarization value.")

    # plural (a bit redundant, but kept the same interfacs as wavefront 1D)
    def get_interpolated_complex_amplitudes(self, x_value,y_value, polarization=Polarization.SIGMA):
        return self.get_interpolated_complex_amplitude(x_value, y_value, polarization=polarization)

    def get_interpolated_amplitudes(self, x_value,y_value, polarization=Polarization.SIGMA): # plural!
        return self.get_interpolated_amplitude(x_value,y_value, polarization=polarization)

    def get_interpolated_phases(self, x_value,y_value, polarization=Polarization.SIGMA): # plural!
        return self.get_interpolated_phase(x_value,y_value, polarization=polarization)

    def get_interpolated_intensities(self, x_value,y_value, polarization=Polarization.SIGMA):
        return self.get_interpolated_intensity(x_value,y_value, polarization=polarization)

    def get_interpolated(self,x_value,y_value,toreturn='complex_amplitude'):
        if toreturn == 'complex_amplitude':
            return self.get_interpolated_amplitudes()
        elif toreturn == 'amplitude':
            return self.get_interpolated_amplitudes()
        elif toreturn == 'phase':
            return self.get_interpolated_phases()
        elif toreturn == 'intensity':
            return self.get_interpolated_intensities()
        else:
            raise Exception('Unknown return string')


    # modifiers

    def set_wavelength(self,wavelength):
        self._wavelength = wavelength

    def set_wavenumber(self,wavenumber):
        self._wavelength = 2*numpy.pi / wavenumber

    def set_photon_energy(self,photon_energy):
        m2ev = codata.c * codata.h / codata.e      # lambda(m)  = m2eV / energy(eV)
        self._wavelength = m2ev / photon_energy

    def set_complex_amplitude(self,complex_amplitude, complex_amplitude_pi=None):
        if self._electric_field_matrix.shape() != complex_amplitude.shape:
            raise Exception("Incompatible shape")
        self._electric_field_matrix.set_z_values(complex_amplitude)

        if complex_amplitude_pi is not None:
            if self.is_polarized():
                if self._electric_field_matrix_pi.shape() != complex_amplitude_pi.shape:
                    raise Exception("Incompatible shape")
                self._electric_field_matrix_pi.set_z_values(complex_amplitude_pi)
            else:
                raise Exception('Cannot set PI-polarized complex amplitude to a non-polarized wavefront.')


    def set_pi_complex_amplitude_to_zero(self):
        if self.is_polarized():
            new_value = self._electric_field_matrix_pi.get_z_values()
            self._electric_field_matrix_pi.set_z_values(new_value*0.0)

    # TODO: add inclination like for 1D
    def set_plane_wave_from_complex_amplitude(self, complex_amplitude=(1.0 + 0.0j)):
        new_value = self._electric_field_matrix.get_z_values()
        new_value *= 0.0
        new_value += complex_amplitude
        self._electric_field_matrix.set_z_values(new_value)
        # if polarized, set arbitrary PI component to zero
        self.set_pi_complex_amplitude_to_zero()


    # TODO: add inclination like for 1D
    def set_plane_wave_from_amplitude_and_phase(self, amplitude=1.0, phase=0.0):
        self.set_plane_wave_from_complex_amplitude(amplitude*numpy.cos(phase) + 1.0j*amplitude*numpy.sin(phase))
        # if polarized, set arbitrary PI component to zero
        self.set_pi_complex_amplitude_to_zero()

    # TODO: add center like for 1D
    def set_spherical_wave(self,  radius=1.0, complex_amplitude=1.0,):
        """

        :param radius:  Positive radius is divergent wavefront, negative radius is convergent
        :param complex_amplitude:
        :return:
        """
        if radius == 0:
            raise Exception("Radius cannot be zero")
        new_value = complex_amplitude * numpy.exp( -1.0j * self.get_wavenumber() *
                                (self.get_mesh_x()**2 + self.get_mesh_y()**2) / (-2*radius) )

        # new_value = numpy.exp(-1.0j * self.get_wavenumber() *
        #                         (self.get_mesh_x()**2+self.get_mesh_y()**2)/(-2*radius))
        self._electric_field_matrix.set_z_values(new_value)
        # if polarized, set arbitrary PI component to zero
        self.set_pi_complex_amplitude_to_zero()

    def set_gaussian_hermite_mode(self, sigma_x, sigma_y, nx, ny, amplitude=1.0, center_x=0.0, center_y=0.0,
                                  betax=100.0,betay=100):
        x = self.get_coordinate_x()
        y = self.get_coordinate_y()

        a2D = GaussianSchellModel2D(amplitude, sigma_x, betax*sigma_x, sigma_y, betay*sigma_y)
        Phi = a2D.phi_nm(nx, ny, x-center_x, y-center_y) + 0j
        eigenvalue = a2D.beta(nx,ny)

        self.set_complex_amplitude(numpy.sqrt(eigenvalue)*Phi)
        # if polarized, set arbitrary PI component to zero
        self.set_pi_complex_amplitude_to_zero()

    # note that amplitude is for "amplitude" not for intensity!
    def set_gaussian(self, sigma_x, sigma_y, amplitude=1.0, center_x=0.0, center_y=0.0):
        self.set_gaussian_hermite_mode(sigma_x, sigma_y, 0, 0, amplitude=amplitude, center_x=center_x, center_y=center_y)
        # if polarized, set arbitrary PI component to zero
        self.set_pi_complex_amplitude_to_zero()


    def add_phase_shift(self, phase_shift, polarization=Polarization.SIGMA):
        if polarization == Polarization.SIGMA:
            new_value = self._electric_field_matrix.get_z_values()
            new_value *= numpy.exp(1.0j*phase_shift)
            self._electric_field_matrix.set_z_values(new_value)
        elif polarization == Polarization.PI:
            if self.is_polarized():
                new_value = self._electric_field_matrix_pi.get_z_values()
                new_value *= numpy.exp(1.0j*phase_shift)
                self._electric_field_matrix_pi.set_z_values(new_value)
            else:
                raise Exception("Wavefront is not polarized")
        else:
            raise Exception("Invalid polarization value (only 0=SIGMA or 1=PI are valid)")

    def add_phase_shifts(self, phase_shifts, polarization=Polarization.SIGMA):

        if polarization == Polarization.SIGMA:
            if phase_shifts.shape != self._electric_field_matrix.shape():
                raise Exception("Phase Shifts array has different dimension")
            new_value = self._electric_field_matrix.get_z_values()
            new_value *= numpy.exp(1.0j*phase_shifts)
            self._electric_field_matrix.set_z_values(new_value)
        elif polarization == Polarization.PI:
            if self.is_polarized():
                if phase_shifts.shape != self._electric_field_matrix_pi.shape():
                    raise Exception("Phase Shifts array has different dimension")
                new_value = self._electric_field_matrix_pi.get_z_values()
                new_value *= numpy.exp(1.0j*phase_shifts)
                self._electric_field_matrix_pi.set_z_values(new_value)
            else:
                raise Exception("Wavefront is not polarized")
        else:
            raise Exception("Invalid polarization value (only 0=SIGMA or 1=PI are valid)")

        # if phase_shifts.shape != self._electric_field_matrix.shape():
        #     raise Exception("Phase Shifts array has different dimension")
        # new_value = self._electric_field_matrix.get_z_values()
        # new_value *= numpy.exp(1.0j*phase_shifts)
        # self._electric_field_matrix.set_z_values(new_value)

    def rescale_amplitude(self, factor, polarization=Polarization.SIGMA):

        if polarization == Polarization.SIGMA:
            new_value = self._electric_field_matrix.get_z_values()
            new_value *= factor
            self._electric_field_matrix.set_z_values(new_value)
        elif polarization == Polarization.PI:
            if self.is_polarized():
                new_value = self._electric_field_matrix_pi.get_z_values()
                new_value *= factor
                self._electric_field_matrix_pi.set_z_values(new_value)
            else:
                raise Exception("Wavefront is not polarized")
        elif polarization == Polarization.TOTAL:
            self.rescale_amplitude(factor, polarization=Polarization.SIGMA)
            self.rescale_amplitude(factor, polarization=Polarization.PI)
        else:
            raise Exception("Invalid polarization value (only 0=SIGMA, 1=PI or 3=TOTAL are valid)")

    def rescale_amplitudes(self, factors, polarization=Polarization.SIGMA):

        if polarization == Polarization.SIGMA:
            if factors.shape != self._electric_field_matrix.shape():
                raise Exception("Factors array has different dimension")
            new_value = self._electric_field_matrix.get_z_values()
            new_value *= factors
            self._electric_field_matrix.set_z_values(new_value)
        elif polarization == Polarization.PI:
            if self.is_polarized():
                if factors.shape != self._electric_field_matrix_pi.shape():
                    raise Exception("Factors array has different dimension")
                new_value = self._electric_field_matrix_pi.get_z_values()
                new_value *= factors
                self._electric_field_matrix_pi.set_z_values(new_value)
            else:
                raise Exception("Wavefront is not polarized")
        elif polarization == Polarization.TOTAL:
            self.rescale_amplitudes(factors, polarization=Polarization.SIGMA)
            self.rescale_amplitudes(factors, polarization=Polarization.PI)
        else:
            raise Exception("Invalid polarization value (only 0=SIGMA, 1=PI or 3=TOTAL are valid)")

    def rebin(self,expansion_points_horizontal, expansion_points_vertical, expansion_range_horizontal, expansion_range_vertical,
              keep_the_same_intensity=0,set_extrapolation_to_zero=0):

        x0 = self.get_coordinate_x()
        y0 = self.get_coordinate_y()

        x1 = numpy.linspace(x0[0]*expansion_range_horizontal,x0[-1]*expansion_range_horizontal,int(x0.size*expansion_points_horizontal))
        y1 = numpy.linspace(y0[0]*expansion_range_vertical  ,y0[-1]*expansion_range_vertical,  int(y0.size*expansion_points_vertical))

        X1 = numpy.outer(x1,numpy.ones_like(y1))
        Y1 = numpy.outer(numpy.ones_like(x1),y1)

        z1 = self.get_interpolated_complex_amplitudes(X1,Y1)

        if set_extrapolation_to_zero:
            z1[numpy.where( X1 < x0[0])] = 0.0
            z1[numpy.where( X1 > x0[-1])] = 0.0
            z1[numpy.where( Y1 < y0[0])] = 0.0
            z1[numpy.where( Y1 > y0[-1])] = 0.0


        if keep_the_same_intensity:
            if self.get_intensity(polarization=Polarization.SIGMA).sum() > 1e-10:
                z1 /= numpy.sqrt( (numpy.abs(z1)**2).sum() / self.get_intensity(polarization=Polarization.SIGMA).sum() )
        #
        if self.is_polarized():
            z1_pi = self.get_interpolated_complex_amplitudes(X1,Y1,polarization=Polarization.PI)

            if set_extrapolation_to_zero:
                z1_pi[numpy.where( X1 < x0[0])] = 0.0
                z1_pi[numpy.where( X1 > x0[-1])] = 0.0
                z1_pi[numpy.where( Y1 < y0[0])] = 0.0
                z1_pi[numpy.where( Y1 > y0[-1])] = 0.0

            if keep_the_same_intensity:
                if self.get_intensity(polarization=Polarization.PI).sum() > 1e-10:
                    z1_pi /= numpy.sqrt( (numpy.abs(z1_pi)**2).sum() / self.get_intensity(polarization=Polarization.PI).sum() )
        else:
            z1_pi = None

        z1_pi = None
        new_wf = GenericWavefront2D.initialize_wavefront_from_arrays(x1,y1,z1,z1_pi,wavelength=self.get_wavelength())

        return new_wf

    def clip_window(self,window):

        if self.is_polarized():
            self.rescale_amplitude(window,polarization=Polarization.TOTAL)
        else:
            self.rescale_amplitudes(window,polarization=Polarization.SIGMA)

    # todo: rename to rectangle!
    def clip_square(self, x_min, x_max, y_min, y_max, negative=False, apply_to_wavefront=True):

        if not negative:
            window = numpy.ones(self._electric_field_matrix.shape())
            lower_window_x = numpy.where(self.get_coordinate_x() < x_min)
            upper_window_x = numpy.where(self.get_coordinate_x() > x_max)
            lower_window_y = numpy.where(self.get_coordinate_y() < y_min)
            upper_window_y = numpy.where(self.get_coordinate_y() > y_max)

            if len(lower_window_x) > 0: window[lower_window_x,:] = 0
            if len(upper_window_x) > 0: window[upper_window_x,:] = 0
            if len(lower_window_y) > 0: window[:,lower_window_y] = 0
            if len(upper_window_y) > 0: window[:,upper_window_y] = 0
        else:
            window = numpy.ones(self._electric_field_matrix.shape())
            window2 = numpy.ones_like(window)
            window_x = numpy.where( (x_min <= self.get_coordinate_x()) & ( self.get_coordinate_x() <= x_max))
            window_y = numpy.where( (y_min <= self.get_coordinate_y()) & ( self.get_coordinate_y() <= y_max))

            if len(window_x) > 0: window[window_x,:] = 0.0
            if len(window_y) > 0: window2[:,window_y] = 0.0

            window += window2
            window_good = numpy.where( window > 0 )
            if len(window_good) > 0: window[window_good] = 1.0

        if apply_to_wavefront:
            self.clip_window(window)

        return window


    def clip_circle(self, radius, x_center=0.0, y_center=0.0, negative=False, apply_to_wavefront=True):
        window = numpy.zeros(self._electric_field_matrix.shape())
        X = self.get_mesh_x()
        Y = self.get_mesh_y()
        distance_to_center = numpy.sqrt( (X-x_center)**2 + (Y-y_center)**2 )
        if negative:
            indices_good = numpy.where(distance_to_center >= radius)
        else:
            indices_good = numpy.where(distance_to_center <= radius)
        window[indices_good] = 1.0

        if apply_to_wavefront:
            self.clip_window(window)

        return window

    def clip_ellipse(self, axis_a, axis_b, x_center=0.0, y_center=0.0, negative=False, apply_to_wavefront=True):
        window = numpy.zeros(self._electric_field_matrix.shape())
        X = self.get_mesh_x()
        Y = self.get_mesh_y()

        TESTX= (X - x_center)**2 / (axis_a/2)**2 + (Y - y_center)**2 / (axis_b/2)**2 - 1.0
        TESTX *= -1.0

        if negative:
            indices_good = numpy.where(TESTX  <= 0.0)
        else:
            indices_good = numpy.where(TESTX >= 0.0)
        window[indices_good] = 1.0

        if apply_to_wavefront:
            self.clip_window(window)

        return window


    def is_identical(self,wfr,decimal=7):
        from numpy.testing import assert_array_almost_equal
        try:
            assert_array_almost_equal(self.get_complex_amplitude(),wfr.get_complex_amplitude(),decimal)
            assert(self.is_polarized() == wfr.is_polarized())
            if self.is_polarized():
                assert_array_almost_equal(self.get_complex_amplitude(polarization=Polarization.PI),
                                          wfr.get_complex_amplitude(polarization=Polarization.PI),decimal)
            assert_array_almost_equal(self.get_coordinate_x(),wfr.get_coordinate_x(),decimal)
            assert_array_almost_equal(self.get_coordinate_y(),wfr.get_coordinate_y(),decimal)
            assert_array_almost_equal(self.get_photon_energy(),wfr.get_photon_energy(),decimal)
        except:
            return False

        return True

    # auxiliary function to dump h5 files
    def _dump_arr_2_hdf5(self,_arr,_calculation, _filename, _subgroupname):
        """
        Auxiliary routine to save_h5_file
        :param _arr: (usually 2D) array to be saved on the hdf5 file inside the _subgroupname
        :param _calculation
        :param _filename: path to file for saving the wavefront
        :param _subgroupname: container mechanism by which HDF5 files are organised
        """
        sys.stdout.flush()
        f = h5py.File(_filename, 'a')
        try:
            f1 = f.create_group(_subgroupname)
        except:
            f1 = f[_subgroupname]
        fdata = f1.create_dataset(_calculation, data=_arr)
        f.close()

    def save_h5_file(self,filename,subgroupname="wfr",intensity=True,phase=False,overwrite=True,verbose=False):
        """
        Auxiliary function to write wavefront data into a hdf5 generic file.
        When using the append mode to write h5 files, overwriting does not work and makes the code crash. To avoid this
        issue, try/except is used. If by any chance a file should be overwritten, it is firstly deleted and re-written.
        :param self: input / output resulting Wavefront structure (instance of GenericWavefront2D);
        :param filename: path to file for saving the wavefront
        :param subgroupname: container mechanism by which HDF5 files are organised
        :param intensity: writes intensity for sigma and pi polarisation (default=True)
        :param amplitude:
        :param phase:
        :param overwrite: flag that should always be set to True to avoid infinity loop on the recursive part of the function.
        :param verbose: if True, print some file i/o messages
        """
        try:
            if not os.path.isfile(filename):  # if file doesn't exist, create it.
                sys.stdout.flush()
                f = h5py.File(filename, 'w')
                # point to the default data to be plotted
                f.attrs['default']          = 'entry'
                # give the HDF5 root some more attributes
                f.attrs['file_name']        = filename
                f.attrs['file_time']        = time.time()
                f.attrs['creator']          = 'oasys-wofry'
                f.attrs['HDF5_Version']     = h5py.version.hdf5_version
                f.attrs['h5py_version']     = h5py.version.version
                f.close()

            # always writes complex amplitude
            x_polarization = self.get_complex_amplitude(polarization=Polarization.SIGMA)       # sigma
            self._dump_arr_2_hdf5(x_polarization.T, "wfr_complex_amplitude_s", filename, subgroupname)

            if self.is_polarized():
                y_polarization = self.get_complex_amplitude(polarization=Polarization.PI)       # pi
                self._dump_arr_2_hdf5(y_polarization.T, "wfr_complex_amplitude_p", filename, subgroupname)


            if intensity:
                if self.is_polarized():
                    self._dump_arr_2_hdf5(self.get_intensity(polarization=Polarization.TOTAL).T,"wfr_intensity",   filename, subgroupname+"/intensity")
                    self._dump_arr_2_hdf5(self.get_intensity(polarization=Polarization.SIGMA).T,"wfr_intensity_s", filename, subgroupname+"/intensity")
                    self._dump_arr_2_hdf5(self.get_intensity(polarization=Polarization.PI).T,"wfr_intensity_p",    filename, subgroupname+"/intensity")
                else:
                    self._dump_arr_2_hdf5(self.get_intensity(polarization=Polarization.SIGMA).T,"wfr_intensity",   filename, subgroupname+"/intensity")

            if phase:
                if self.is_polarized():
                    self._dump_arr_2_hdf5( \
                        self.get_phase(polarization=Polarization.SIGMA).T - self.get_phase(polarization=Polarization.PI).T,
                                        "wfr_phase", filename, subgroupname+"/phase")
                    self._dump_arr_2_hdf5(self.get_phase(polarization=Polarization.SIGMA).T,"wfr_phase_s", filename, subgroupname+"/phase")
                    self._dump_arr_2_hdf5(self.get_phase(polarization=Polarization.PI).T,"wfr_phase_p", filename, subgroupname+"/phase")
                else:
                    self._dump_arr_2_hdf5(self.get_phase(polarization=Polarization.SIGMA).T,"wfr_phase",   filename, subgroupname+"/phase")



            # add mesh and SRW information




            # add mesh and SRW information


            f = h5py.File(filename, 'a')
            f1 = f[subgroupname]

            # point to the default data to be plotted
            f1.attrs['NX_class'] = 'NXentry'
            f1.attrs['default'] = 'intensity'

            # TODO: add self interpreting decoder
            # f1["wfr_method"] = "WOFRY"
            f1["wfr_dimension"] = 2
            f1["wfr_photon_energy"] = self.get_photon_energy()
            x = self.get_coordinate_x()
            y = self.get_coordinate_y()

            f1["wfr_mesh_X"] =  numpy.array([x[0],x[-1],x.size])
            f1["wfr_mesh_Y"] =  numpy.array([y[0],y[-1],y.size])

            # Add NX plot attribites for automatic plot with silx view
            myflags = [intensity,phase]
            mylabels = ['intensity','phase']
            for i,label in enumerate(mylabels):
                if myflags[i]:
                    f2 = f1[mylabels[i]]
                    f2.attrs['NX_class'] = 'NXdata'
                    f2.attrs['signal'] = 'wfr_%s'%(mylabels[i])
                    f2.attrs['axes'] = [b'axis_y', b'axis_x']

                    f3 = f2["wfr_%s"%(mylabels[i])]
                    f3.attrs['interpretation'] = 'image'

                    # X axis data
                    ds = f2.create_dataset('axis_y', data=1e6*y)
                    ds.attrs['units'] = 'microns'
                    ds.attrs['long_name'] = 'Y Pixel Size (microns)'    # suggested X axis plot label

                    # Y axis data
                    ds = f2.create_dataset('axis_x', data=1e6*x)
                    ds.attrs['units'] = 'microns'
                    ds.attrs['long_name'] = 'X Pixel Size (microns)'    # suggested Y axis plot label
            f.close()

        except:
            # TODO: check exit??
            if overwrite is not True:
                raise Exception("Bad input argument")
            os.remove(filename)
            if verbose: print("save_h5_file: file deleted %s"%filename)
            self.save_h5_file(filename,subgroupname, intensity=intensity, phase=phase, overwrite=False)

        if verbose: print("save_h5_file: witten/updated %s data in file: %s"%(subgroupname,filename))

    @classmethod
    def load_h5_file(cls,filename,filepath="wfr"):

        try:
            f = h5py.File(filename, 'r')
            mesh_X = f[filepath+"/wfr_mesh_X"][()]
            mesh_Y = f[filepath+"/wfr_mesh_Y"][()]
            complex_amplitude_s = f[filepath+"/wfr_complex_amplitude_s"][()].T
            try:
                complex_amplitude_p = f[filepath + "/wfr_complex_amplitude_p"][()].T
            except:
                complex_amplitude_p = None
            wfr = cls.initialize_wavefront_from_arrays(
                                x_array=numpy.linspace(mesh_X[0],mesh_X[1],int(mesh_X[2])),
                                y_array=numpy.linspace(mesh_Y[0],mesh_Y[1],int(mesh_Y[2])),
                                z_array=complex_amplitude_s,
                                z_array_pi=complex_amplitude_p)
            wfr.set_photon_energy(f[filepath+"/wfr_photon_energy"][()])
            f.close()
            return wfr
        except:
            raise Exception("Failed to load 2D wavefront to h5 file: "+filename)

if __name__ == "__main__":
    w = GenericWavefront2D.initialize_wavefront_from_steps(polarization=Polarization.TOTAL)
    w.save_h5_file("/tmp/wf.h5",subgroupname="wfr",intensity=True,phase=False,overwrite=True,verbose=True)
    w2 = GenericWavefront2D.load_h5_file("/tmp/wf.h5",filepath="wfr")
    assert(w2.is_identical(w))
    pass



import unittest
import numpy


#
# Note that the tests for the Fraunhofer phase do not make any assert, because a good matching has not yet been found.
#


from syned.beamline.shape import Rectangle, Ellipse
from syned.beamline.element_coordinates import ElementCoordinates
from syned.beamline.beamline_element import BeamlineElement

from wofry.propagator.propagator import PropagationElements

from wofry.propagator.wavefront1D.generic_wavefront import GenericWavefront1D
from wofry.propagator.propagator import PropagationManager, PropagationParameters

from wofry.beamline.optical_elements.absorbers.slit import WOSlit1D, WOGaussianSlit1D


do_plot = False

if do_plot:
    from srxraylib.plot.gol import plot

from wofry.propagator.propagators1D.fresnel_zoom import FresnelZoom1D

from wofry.propagator.propagators1D.fraunhofer import Fraunhofer1D
from wofry.propagator.propagators1D.fresnel import Fresnel1D
from wofry.propagator.propagators1D.fresnel_convolution import FresnelConvolution1D
from wofry.propagator.propagators1D.integral import Integral1D
from wofry.propagator.propagators1D import initialize_default_propagator_1D

from scipy.special import fresnel



propagator = PropagationManager.Instance()
initialize_default_propagator_1D()


def fraunhofer_analytical_rectangle(
    fresnel_number=None,propagation_distance=1.140,
    aperture_half=1e-3,wavelength=639e-9,
    detector_array=None,npoints=1000,
    ):


    if fresnel_number is None:
        fresnel_number = aperture_half**2 / (wavelength * propagation_distance)

    print("Fresnel number: ",fresnel_number)


    if detector_array is None:
        if fresnel_number > 1.0:
            window_aperture_ratio = 2.0
        else:
            window_aperture_ratio = 1.0 / fresnel_number
        x = numpy.linspace(-window_aperture_ratio*aperture_half,window_aperture_ratio*aperture_half,npoints)
    else:
        x = detector_array.copy()

    argument_sinc = 2.0 * aperture_half * numpy.pi / wavelength / propagation_distance * x # TODO: check the 2??
    alpha = 2.0 * aperture_half / (wavelength*propagation_distance)**(1.0/2.0) * \
            numpy.exp(1j*numpy.pi/wavelength/propagation_distance * x**2) * \
            numpy.sin(argument_sinc) / argument_sinc

    # TODO note that the global phase (Goldman 4-59) is missing

    return x,alpha

def fresnel_analytical_rectangle(
    fresnel_number=None,propagation_distance=1.140,
    aperture_half=1e-3,wavelength=639e-9,
    detector_array=None,npoints=1000,
    ):


    if fresnel_number is None:
        fresnel_number = aperture_half**2 / (wavelength * propagation_distance)

    print("Fresnel number: ",fresnel_number)


    if detector_array is None:
        if fresnel_number > 1.0:
            window_aperture_ratio = 2.0
        else:
            window_aperture_ratio = 1.0 / fresnel_number
        x = numpy.linspace(-window_aperture_ratio*aperture_half,window_aperture_ratio*aperture_half,npoints)
    else:
        x = detector_array.copy()

    s_plus  = numpy.sqrt(2.0 * fresnel_number) * ( 1.0 + x / aperture_half)
    s_minus = numpy.sqrt(2.0 * fresnel_number) * ( 1.0 - x / aperture_half)

    fs_plus,fc_plus = fresnel(s_plus)
    fs_minus,fc_minus = fresnel(s_minus)

    Ux = (fc_minus + fc_plus) + 1j*(fs_minus + fs_plus)
    Ux *= 1.0/numpy.sqrt(2.0)

    # TODO note that the global phase (Goldman 4-59) is missing

    return x,Ux  # note that wavefield is being returned, not intensity!


class propagatorTest(unittest.TestCase):

    #
    # Common interface for all 1D methods :
    #                                   'fraunhofer':
    #                                   'fft': fft -> multiply by kernel in freq -> ifft
    #                                   'convolution': scipy.signal.fftconvolve(wave,kernel in space)
    #                                   'zoom': fft -> multiply by kernel in freq -> ifft
    # valid apertute_type: square, gaussian

    def propagate_1D(self,do_plot=do_plot,
                                method='fft',
                                wavelength=1.24e-10,
                                aperture_type='square',
                                aperture_diameter=40e-6,
                                wavefront_length=100e-6,
                                npoints=500,
                                propagation_distance = 30.0,
                                normalization=True,  # TODO put False
                                show=1,
                                amplitude=(0.0+1.0j)):



        print("\n#                                                            ")
        print("# 1D (%s) propagation from a %s aperture  "%(method,aperture_type))
        print("#                                                            ")

        wf = GenericWavefront1D.initialize_wavefront_from_range(x_min=-wavefront_length/2, x_max=wavefront_length/2,
                                                                number_of_points=npoints,wavelength=wavelength)

        wf.set_plane_wave_from_complex_amplitude(amplitude) # an arbitraty value

        deltax = wf.get_abscissas()[1] - wf.get_abscissas()[0]

        propagation_elements = PropagationElements()

        slit = None

        if aperture_type == 'square':
            slit = WOSlit1D(boundary_shape=Rectangle(-aperture_diameter/2, aperture_diameter/2, 0, 0))
        elif aperture_type == 'gaussian':
            slit = WOGaussianSlit1D(boundary_shape=Rectangle(-aperture_diameter/2, aperture_diameter/2, 0, 0))
        else:
            raise Exception("Not implemented! (accepted: circle, square, gaussian)")

        propagation_elements.add_beamline_element(BeamlineElement(optical_element=slit,
                                                                  coordinates=ElementCoordinates(p=0, q=propagation_distance)  ))


        propagator = PropagationManager.Instance()
        propagation_parameters = PropagationParameters(wavefront=wf,
                                                       propagation_elements=propagation_elements)

        print("Using propagator method:  ",method)
        fresnel_analytical = True
        if method == 'fft':
            wf1 = propagator.do_propagation(propagation_parameters, Fresnel1D.HANDLER_NAME)
        elif method == 'convolution':
            wf1 = propagator.do_propagation(propagation_parameters, FresnelConvolution1D.HANDLER_NAME)
        elif method == 'integral':
            propagation_parameters.set_additional_parameters("magnification_x", 1.5)
            propagation_parameters.set_additional_parameters("magnification_N", 2.0)
            wf1 = propagator.do_propagation(propagation_parameters, Integral1D.HANDLER_NAME)
        elif method == 'fraunhofer':
            fresnel_analytical = False
            # propagation_parameters.set_additional_parameters("shift_half_pixel", 0)
            wf1 = propagator.do_propagation(propagation_parameters, Fraunhofer1D.HANDLER_NAME)
        elif method == 'zoom':
            propagation_parameters.set_additional_parameters("magnification_x", 1.5)
            wf1 = propagator.do_propagation(propagation_parameters, FresnelZoom1D.HANDLER_NAME)
        else:
            raise Exception("Not implemented method: %s"%method)


        if fresnel_analytical:
            xx, alpha = fresnel_analytical_rectangle(
                fresnel_number = None, propagation_distance = propagation_distance,
                aperture_half = 0.5*aperture_diameter, wavelength = wavelength,
                detector_array = wf1.get_abscissas(), npoints = None)
        else:
            xx, alpha = fraunhofer_analytical_rectangle(
                fresnel_number = None, propagation_distance = propagation_distance,
                aperture_half = 0.5*aperture_diameter, wavelength = wavelength,
                detector_array = wf1.get_abscissas(), npoints = None)

        angle_x = xx / propagation_distance
        intensity_theory = numpy.abs(amplitude*alpha)**2

        intensity_calculated =  wf1.get_intensity()

        if normalization:
            intensity_calculated /= intensity_calculated.max()
            intensity_theory /= intensity_theory.max()

        if do_plot:
            from srxraylib.plot.gol import plot
            plot(wf1.get_abscissas()*1e6/propagation_distance,intensity_calculated,
                 angle_x*1e6,intensity_theory,
                 legend=["%s "%method,"analytical"],
                 legend_position=(0.95, 0.95),
                 title="1D (%s) diffraction from a %s aperture of %3.1f um at \n wavelength of %3.1f A"%
                       (method,aperture_type,aperture_diameter*1e6,wavelength*1e10),
                 xtitle="X (urad)", ytitle="Intensity",
                 xrange=[-20,20],ylog=True,
                 show=show)

            plot(wf1.get_abscissas()*1e6,wf1.get_phase(unwrap=True),
                 1e6 * xx, numpy.unwrap(numpy.angle(alpha)),
                 legend=["%s " % method, "analytical"],
                 title="1D (%s) diffraction from a %s aperture of %3.1f um at \n wavelength of %3.1f A NOT ASSERTED!!"%
                       (method, aperture_type, aperture_diameter * 1e6, wavelength * 1e10),
                 xtitle="X (urad)", ytitle="Phase",
                 # xrange=[-20, 20],
                 )

        return wf1.get_abscissas()/propagation_distance,intensity_calculated,intensity_theory


    #
    # tests/example cases for 1D propagators
    #


    # @unittest.skip("classing skipping")
    def test_propagate_1D_fraunhofer_bis(self,do_plot=do_plot):

        aperture_type="square"
        aperture_diameter = 40e-6
        wavefront_length = 800e-6
        wavelength = 1.24e-10
        npoints=1024

        print("\n#                                                            ")
        print("# far field 1D (fraunhofer) diffraction from a %s aperture  "%aperture_type)
        print("#                                                            ")



        angle, intensity_calculated,intensity_theory = self.propagate_1D(do_plot=do_plot,method="fraunhofer",
                                wavelength=wavelength,aperture_type=aperture_type,aperture_diameter=aperture_diameter,
                                wavefront_length=wavefront_length,npoints=npoints,
                                propagation_distance = 1.0, show=1)

        # works for normalized only
        numpy.testing.assert_almost_equal(intensity_calculated,intensity_theory,1)

    # @unittest.skip("classing skipping")
    def test_propagate_1D_fraunhofer_phase(self,do_plot=do_plot):

        # aperture_type="square"
        # aperture_diameter = 40e-6
        # wavefront_length = 800e-6
        # wavelength = 1.24e-10
        # npoints=1024
        # propagation_distance=40
        show = 1

        aperture_type="square"
        aperture_diameter = 40e-6
        wavefront_length = 800e-6
        wavelength = 1.24e-10
        propagation_distance = 30.0
        npoints=1024


        print("\n#                                                            ")
        print("# far field 1D (fraunhofer and zoom) diffraction from a %s aperture  "%aperture_type)
        print("#                                                            ")

        wf = GenericWavefront1D.initialize_wavefront_from_range(x_min=-wavefront_length/2, x_max=wavefront_length/2,
                                                                number_of_points=npoints,wavelength=wavelength)

        wf.set_plane_wave_from_complex_amplitude((2.0+1.0j)) # an arbitraty value

        propagation_elements = PropagationElements()

        if aperture_type == 'square':
            slit = WOSlit1D(boundary_shape=Rectangle(-aperture_diameter/2, aperture_diameter/2, 0, 0))
        else:
            raise Exception("Not implemented! ")

        propagation_elements.add_beamline_element(BeamlineElement(optical_element=slit,
                                                                  coordinates=ElementCoordinates(p=0, q=propagation_distance)  ))


        propagator = PropagationManager.Instance()
        propagation_parameters = PropagationParameters(wavefront=wf,
                                                       propagation_elements=propagation_elements)



        wf1_franuhofer = propagator.do_propagation(propagation_parameters, Fraunhofer1D.HANDLER_NAME)

        propagation_parameters.set_additional_parameters("shift_half_pixel", True)
        propagation_parameters.set_additional_parameters("magnification_x", 1.5)
        wf1_zoom = propagator.do_propagation(propagation_parameters, FresnelZoom1D.HANDLER_NAME)


        intensity_fraunhofer = wf1_franuhofer.get_intensity() / wf1_franuhofer.get_intensity().max()
        intensity_zoom = wf1_zoom.get_intensity() / wf1_zoom.get_intensity().max()

        if do_plot:
            from srxraylib.plot.gol import plot
            plot(wf1_franuhofer.get_abscissas()*1e6/propagation_distance,intensity_fraunhofer,
                 wf1_zoom.get_abscissas()*1e6/propagation_distance,intensity_zoom,
                 legend=["Fraunhofer","Zoom"],
                 legend_position=(0.95, 0.95),
                 title="1D  INTENSITY diffraction from aperture of %3.1f um at wavelength of %3.1f A"%
                       (aperture_diameter*1e6,wavelength*1e10),
                 xtitle="X (urad)", ytitle="Intensity",xrange=[-20,20],
                 show=show)
            plot(wf1_franuhofer.get_abscissas()*1e6/propagation_distance,wf1_franuhofer.get_phase(unwrap=1),
                 wf1_zoom.get_abscissas()*1e6/propagation_distance,wf1_zoom.get_phase(unwrap=1),
                 legend=["Fraunhofer","Zoom"],
                 legend_position=(0.95, 0.95),
                 title="1D  diffraction from a %s aperture of %3.1f um at wavelength of %3.1f A"%
                       (aperture_type,aperture_diameter*1e6,wavelength*1e10),
                 xtitle="X (urad)", ytitle="Intensity",xrange=[-20,20],
                 show=show)

        # TODO assert phase
        #numpy.testing.assert_almost_equal(1e3*intensity_fraunhofer,1e3*intensity_zoom,1)

    # @unittest.skip("classing skipping")
    def test_propagate_1D_fft(self,do_plot=do_plot):

        aperture_type="square"
        aperture_diameter = 40e-6
        wavefront_length = 800e-6
        wavelength = 1.24e-10
        propagation_distance = 30.0
        npoints=1024

        # print("\n#                                                            ")
        # print("# near field 1D (fft) diffraction from a %s aperture  "%aperture_type)
        # print("#                                                            ")



        angle, intensity_calculated,intensity_theory = self.propagate_1D(do_plot=do_plot,method="fft",
                                wavelength=wavelength,aperture_type=aperture_type,aperture_diameter=aperture_diameter,
                                wavefront_length=wavefront_length,npoints=npoints,
                                propagation_distance = propagation_distance,
                                show=1)

        numpy.testing.assert_almost_equal(intensity_calculated,intensity_theory,1)

    # @unittest.skip("classing skipping")
    def test_propagate_1D_fft_zoom(self,do_plot=do_plot):

        aperture_type="square"
        aperture_diameter = 40e-6
        wavefront_length = 800e-6
        wavelength = 1.24e-10
        propagation_distance = 30.0
        npoints=1024

        print("\n#                                                            ")
        print("# fft zoom diffraction from a %s aperture  "%aperture_type)
        print("#                                                            ")



        angle, intensity_calculated,intensity_theory = self.propagate_1D(do_plot=do_plot,method="zoom",
                                wavelength=wavelength,aperture_type=aperture_type,aperture_diameter=aperture_diameter,
                                wavefront_length=wavefront_length,npoints=npoints,
                                propagation_distance = propagation_distance,
                                show=1)

        numpy.testing.assert_almost_equal(intensity_calculated,intensity_theory,1)

    # @unittest.skip("classing skipping")
    def test_propagate_1D_fresnel_convolution(self,do_plot=do_plot):

        aperture_type="square"
        aperture_diameter = 40e-6
        wavefront_length = 800e-6
        wavelength = 1.24e-10
        propagation_distance = 30.0
        npoints=1024

        print("\n#                                                            ")
        print("# far field 1D (fraunhofer) diffraction from a %s aperture  "%aperture_type)
        print("#                                                            ")



        angle, intensity_calculated,intensity_theory = self.propagate_1D(do_plot=do_plot,method="convolution",
                                wavelength=wavelength,aperture_type=aperture_type,aperture_diameter=aperture_diameter,
                                wavefront_length=wavefront_length,npoints=npoints,
                                propagation_distance = propagation_distance,
                                show=1)

        numpy.testing.assert_almost_equal(intensity_calculated,intensity_theory,1)


    # @unittest.skip("classing skipping")
    def test_propagate_1D_integral(self,do_plot=do_plot):

        aperture_type="square"
        aperture_diameter = 40e-6
        wavefront_length = 800e-6
        wavelength = 1.24e-10
        propagation_distance = 30.0
        npoints=1024

        print("\n#                                                            ")
        print("# near field 1D (integral) diffraction from a %s aperture  "%aperture_type)
        print("#                                                            ")



        angle, intensity_calculated,intensity_theory = self.propagate_1D(do_plot=do_plot,method="integral",
                                wavelength=wavelength,aperture_type=aperture_type,aperture_diameter=aperture_diameter,
                                wavefront_length=wavefront_length,npoints=npoints,
                                propagation_distance = propagation_distance,
                                show=1)

        numpy.testing.assert_almost_equal(intensity_calculated,intensity_theory,1)


    # @unittest.skip("classing skipping")
    def test_propagate_1D_compare_amplitudes(self,do_plot=do_plot):

        # aperture_type="square"
        # aperture_diameter = 40e-6
        # wavefront_length = 800e-6
        # wavelength = 1.24e-10
        # propagation_distance = 30.0
        # npoints=1024

        aperture_type="square"
        aperture_diameter = 100e-6
        wavefront_length = 1000e-6
        wavelength = 1.5e-10
        propagation_distance = 55.0
        npoints=2048


        print("\n#                                                            ")
        print("# near field 1D (integral) diffraction from a %s aperture  "%aperture_type)
        print("#                                                            ")


        #
        # angle, intensity_calculated,intensity_theory_integral = self.propagate_1D(do_plot=do_plot,method="integral",
        #                         wavelength=wavelength,aperture_type=aperture_type,aperture_diameter=aperture_diameter,
        #                         wavefront_length=wavefront_length,npoints=npoints,
        #                         propagation_distance = propagation_distance, show=1)


        # compare with this one, the reference
        # angle_fft, intensity_calculated_fft,intensity_theory = self.propagate_1D(do_plot=False,method="fft",
        #                         wavelength=wavelength,aperture_type=aperture_type,aperture_diameter=aperture_diameter,
        #                         wavefront_length=wavefront_length,npoints=npoints,
        #                         propagation_distance = propagation_distance,
        #                         show=False)


        for method in ["fft","zoom","convolution","integral","fraunhofer"]:

            angle_1, intensity_calculated_1,intensity_theory = self.propagate_1D(
                                    do_plot=do_plot,
                                    method=method,
                                    wavelength=wavelength,aperture_type=aperture_type,
                                    aperture_diameter=aperture_diameter,
                                    wavefront_length=wavefront_length,npoints=npoints,
                                    propagation_distance = propagation_distance,
                                    show=False)


            # x_fft = angle_fft * propagation_distance
            # x_1 = angle_1 * propagation_distance
            #
            # ymax_all = [intensity_calculated_fft.max(),intensity_calculated_1.max()]
            # ymax = numpy.max(ymax_all)
            # print(">>>",ymax_all,numpy.sqrt(intensity_calculated_fft.max()/intensity_calculated_1.max()))
            # if do_plot:
            #     plot(x_fft,intensity_calculated_fft,x_1,intensity_calculated_1,
            #          legend=["fft",method],yrange=[0,ymax],
            #          title="Comparing intensities - Not yet a test!",
            #          ylog=True,
            #          show=True)

            # numpy.testing.assert_almost_equal(intensity_calculated_fft,intensity_calculated_1,1)








# THIS FILE IS GENERATED FROM wofry SETUP.PY
short_version = '1.0.23'
version = '1.0.23'
full_version = '1.0.23'
git_revision = '23c649e9fa948214b2feabe84ea0306fc42638de'
release = True

if not release:
    version = full_version
    short_version += ".dev"



from srxraylib.plot.gol import plot,plot_image


# source
from wofry.propagator.wavefront2D.generic_wavefront import GenericWavefront2D
# beamline
from syned.beamline.element_coordinates import ElementCoordinates
from syned.beamline.beamline_element import BeamlineElement
from wofry.beamline.optical_elements.ideal_elements.screen import WOScreen as Screen
# propagator
from wofry.propagator.propagator import PropagationManager, PropagationParameters
from wofry.propagator.propagator import PropagationElements
from wofry.propagator.propagators2D.fraunhofer import Fraunhofer2D
from wofry.propagator.propagators2D.fresnel import Fresnel2D
from wofry.propagator.propagators2D.fresnel_convolution import FresnelConvolution2D
from wofry.propagator.propagators2D import initialize_default_propagator_2D

propagator = PropagationManager.Instance()
initialize_default_propagator_2D()

# try:
#     from wofry.propagator.test.propagators.srw_fresnel import FresnelSRW
#     propagator.add_propagator(FresnelSRW())
# except:
#     print("FresnelSRW cannot be added")


if __name__ == "__main__":

    wavelength=0.15e-9
    aperture_diameter=50e-6
    pixelsize_x=1e-7
    pixelsize_y=1e-7
    npixels_x=2024
    npixels_y=2024
    propagation_distance = 1.0
    show=1

    method_label = "fresnel (fft)"
    print("\n#                                                             ")
    print("# 2D near field fresnel (%s) diffraction from a a circular stop  "%(method_label))
    print("#                                                             ")


    wf = GenericWavefront2D.initialize_wavefront_from_range(x_min=-pixelsize_x*npixels_x/2,x_max=pixelsize_x*npixels_x/2,
                                                     y_min=-pixelsize_y*npixels_y/2,y_max=pixelsize_y*npixels_y/2,
                                                     number_of_points=(npixels_x,npixels_y),wavelength=wavelength)

    wf.set_plane_wave_from_complex_amplitude((1.0+0j))


    wf.clip_circle(aperture_diameter/2,negative=True)

    plot_image(wf.get_intensity(),1e6*wf.get_coordinate_x(),1e6*wf.get_coordinate_y(),
               title="intensity at screen/aperture plane, Diameter=%5.1f um"%
                     (1e6*aperture_diameter),xtitle="X [um]",ytitle="Y [um]",
               show=0)

    #
    # define image plane
    #
    propagation_elements = PropagationElements()
    #
    propagation_elements.add_beamline_element(BeamlineElement(optical_element=Screen(),
                                                              coordinates=ElementCoordinates(p=0, q=propagation_distance)))
    propagation_parameters = PropagationParameters(wavefront=wf,
                                                   propagation_elements=propagation_elements)
    #
    method = 'fft'
    #
    if method == 'fft':
        propagation_parameters.set_additional_parameters("shift_half_pixel", True)
        wf1 = propagator.do_propagation(propagation_parameters, Fresnel2D.HANDLER_NAME)
    elif method == 'srw':
        # wf1 = propagator.do_propagation(propagation_parameters, FresnelSRW.HANDLER_NAME)
        raise Exception("To be implemented using wofrysrw")
    elif method == 'convolution':
        propagation_parameters.set_additional_parameters("shift_half_pixel", True)
        wf1 = propagator.do_propagation(propagation_parameters, FresnelConvolution2D.HANDLER_NAME)
    elif method == 'fraunhofer':
        propagation_parameters.set_additional_parameters("shift_half_pixel", True)
        wf1 = propagator.do_propagation(propagation_parameters, Fraunhofer2D.HANDLER_NAME)
    else:
        raise Exception("Not implemented method: %s"%method)




    plot_image(wf1.get_intensity(),
               1e6*wf1.get_coordinate_x()/propagation_distance,
               1e6*wf1.get_coordinate_y()/propagation_distance,
               title="Diffracted intensity by a circular stop %3.1f um"%
                     (1e6*aperture_diameter),
               xtitle="X [urad]",ytitle="Y [urad]",
               show=0)


    intensity_calculated =  wf1.get_intensity()[:,int(wf1.size()[1]/2)]

    intensity_calculated /= intensity_calculated.max()

    plot(wf1.get_coordinate_x()*1e6/propagation_distance,intensity_calculated,
         legend=["%s H profile"%method_label],
         legend_position=(0.95, 0.95),
         title="%s diffraction of a cirlular stop %3.1f um at wavelength of %3.1f A"%
               (method_label,aperture_diameter*1e6,wavelength*1e10),
         xtitle="X (urad)", ytitle="Intensity",xrange=[-100,100],
         show=show)





class WavefrontDecorator():

    def get_dimension(self):
        raise NotImplementedError()

    def toGenericWavefront(self):
        raise NotImplementedError()

    @classmethod
    def fromGenericWavefront(cls, wavefront):
        raise NotImplementedError()


import numpy as np
import scipy.special as special

class GaussianSchellModel1D(object):
    def __init__(self, A, sigma_s, sigma_g):
        """
        Mandel and Wolf p 253

        :param A: Amplitude of the spectral density S
        :param sigma_s: Gaussian width of the spectral density S
        :param sigma_g: Gaussian width of spectral degree of coherence g
        """
        self._A = A
        self._sigma_s = sigma_s
        self._sigma_g = sigma_g

    def S(self, x):
        S = (self._A ** 2) * np.exp(-(x**2)/(2*self._sigma_s**2))
        return S

    def g(self, x):
        g = np.exp(-(x**2)/(2*self._sigma_g**2))
        return g

    def evaluate(self, x_1, x_2):

        dr = x_1-x_2

        S_1 = self.S(x_1)
        S_2 = self.S(x_2)
        g = self.g(dr)

        result = np.sqrt(S_1) * np.sqrt(S_2) * g

        return result

    def a(self):
        a = 1.0/(4.0*self._sigma_s**2)
        return a

    def b(self):
        b = 1.0/(2.0*self._sigma_g**2)
        return b

    def c(self):
        a = self.a()
        b = self.b()

        res = np.sqrt(a**2 + 2 * a * b)
        return res

    def beta(self,n):
        a = self.a()
        b = self.b()
        c = self.c()

        res = self._A**2 * np.sqrt(np.pi/(a+b+c)) * (b/(a+b+c))**n
        return res

    def phi(self, n, x):
        c_h_n = special.hermite(n)
        c = self.c()
        h_n = np.polyval(c_h_n, x * np.sqrt(2*c) )

        res = ((2*c/np.pi) ** 0.25) * np.exp(-c * x**2)
        res *= h_n * (1.0/np.sqrt(2**n * special.factorial(n) ))

        return res

class GaussianSchellModel2D(object):
    def __init__(self, A, sigma_s_x, sigma_g_x, sigma_s_y, sigma_g_y):
        """
        Mandel and Wolf p 253

        :param A: Amplitude of the spectral density S
        :param sigma_s: Gaussian width of the spectral density S
        :param sigma_g: Gaussian width of spectral degree of coherence g
        """

        self._mode_x = GaussianSchellModel1D(A**0.5, sigma_s_x, sigma_g_x)
        self._mode_y = GaussianSchellModel1D(A**0.5, sigma_s_y, sigma_g_y)

        # For eigenvalue ordering.
        self._sorted_mode_indices = None

    def evaluate(self, r_1, r_2):
        x = self._mode_x.evaluate(r_1[0], r_2[0])
        y = self._mode_y.evaluate(r_1[1], r_2[1])

        result = x * y

        return result

    def beta(self, n_x, n_y):
        beta_x = self._mode_x.beta(n_x)
        beta_y = self._mode_y.beta(n_y)

        res = beta_x * beta_y

        return res

    def phi(self, n_x, n_y, x, y):
        phi_x = self._mode_x.phi(n_x, x)
        phi_y = self._mode_y.phi(n_y, y)

        res = phi_x * phi_y

        return res

    def phi_nm(self,n_x, n_y, x_coords, y_coords):
        phi_x = self._mode_x.phi(n_x, x_coords)
        phi_y = self._mode_y.phi(n_y, y_coords)

        res = np.outer(phi_x,  phi_y)
        return res

    def sortedModeIndices(self, index_energy, n_points=50):
        if self._sorted_mode_indices is None:
            eigenvalues_x = np.array([self._mode_x.beta(i) for i in range(n_points)])
            eigenvalues_y = np.array([self._mode_y.beta(i) for i in range(n_points)])

            f = np.outer(eigenvalues_x, eigenvalues_y)
            self._sorted_mode_indices = np.array(np.unravel_index(f.flatten().argsort()[::-1], (n_points, n_points)))

        n, m = self._sorted_mode_indices[:, index_energy]
        return n, m

if __name__ == "__main__":

    from srxraylib.plot.gol import plot,plot_image
    # 1D
    sigmaX = 100e-6
    modeX = 0
    a1D = GaussianSchellModel1D(1.0,sigmaX,100.0*sigmaX)

    x = np.linspace(-5*sigmaX,5*sigmaX,100)
    y = a1D.phi(modeX,x)

    plot(x,y)

    print(">>",y[30])

    # 2D
    sigmaX = 100e-6
    modeX = 0
    nX = 100
    sigmaY = 50e-6
    modeY = 3
    nY = 100
    a2D = GaussianSchellModel2D(1.0,sigmaX,100.0*sigmaX,sigmaY,100.0*sigmaY)

    x = np.linspace(-5*sigmaX,5*sigmaX,nX)
    y = np.linspace(-5*sigmaY,5*sigmaY,nY)

    Phi = a2D.phi_nm(modeX,modeY,x,y)

    print(">>",Phi[30,40])

    plot_image(Phi)



import numpy

from syned.beamline.optical_elements.absorbers.slit import Slit
from syned.beamline.shape import BoundaryShape, Rectangle, Circle, Ellipse, MultiplePatch

from wofry.beamline.decorators import OpticalElementDecorator

class WOSlit(Slit, OpticalElementDecorator):
    def __init__(self, name="Undefined", boundary_shape=BoundaryShape()):
        Slit.__init__(self, name=name, boundary_shape=boundary_shape)

    def applyOpticalElement(self, wavefront, parameters=None, element_index=None):
        boundaries = self._boundary_shape.get_boundaries()

        if isinstance(self._boundary_shape, Rectangle):
            wavefront.clip_square(boundaries[0], boundaries[1], boundaries[2], boundaries[3])
        elif isinstance(self._boundary_shape, Circle):
            wavefront.clip_circle(boundaries[0], boundaries[1], boundaries[2])
        elif isinstance(self._boundary_shape, Ellipse):
            #clip_ellipse expects axes and center
            a_axis = boundaries[1] - boundaries[0]
            b_axis = boundaries[3] - boundaries[2]
            x_center = 0.5 * (boundaries[0] + boundaries[1])
            y_center = 0.5 * (boundaries[2] + boundaries[3])
            wavefront.clip_ellipse(a_axis, b_axis, x_center, y_center)
        elif isinstance(self._boundary_shape, MultiplePatch):
            windows = []
            for i in range(self._boundary_shape.get_number_of_patches()):
                patch = self._boundary_shape.get_patch(i)
                bd = patch.get_boundaries()
                if self._boundary_shape.get_name_of_patch(i) == "Rectangle":
                    windows.append( wavefront.clip_square(bd[0],bd[1],bd[2],bd[3],apply_to_wavefront=False) )
                elif self._boundary_shape.get_name_of_patch(i) == "Circle":
                    windows.append( wavefront.clip_circle(bd[0],bd[1],bd[2],apply_to_wavefront=False) )
                elif self._boundary_shape.get_name_of_patch(i) == "Ellipse":
                    print(bd)
                    a_axis = bd[1] - bd[0]
                    b_axis = bd[3] - bd[2]
                    x_center = 0.5 * (bd[0] + bd[1])
                    y_center = 0.5 * (bd[2] + bd[3])
                    windows.append( wavefront.clip_ellipse(a_axis,b_axis,x_center,y_center,apply_to_wavefront=False) )
                else:
                    raise Exception(NotImplementedError)

            final_window = numpy.zeros_like(windows[0])

            for i in range(len(windows)):
                final_window += windows[i]

            final_window[ final_window > 0 ] = 1.0 # renormalize

            wavefront.clip_window(final_window)
        else:
            raise NotImplementedError("to be implemented")

        return wavefront

class WOGaussianSlit(Slit, OpticalElementDecorator):
    def __init__(self, name="Undefined", boundary_shape=BoundaryShape()):
        Slit.__init__(self, name=name, boundary_shape=boundary_shape)

    def applyOpticalElement(self, wavefront, parameters=None, element_index=None):
        boundaries = self._boundary_shape.get_boundaries()
        aperture_diameter_x =  numpy.abs(boundaries[1] - boundaries[0])
        aperture_diameter_y =  numpy.abs(boundaries[2] - boundaries[3])
        X = wavefront.get_mesh_x()
        Y = wavefront.get_mesh_y()

        wavefront.rescale_amplitudes(numpy.exp(- (((X*X)/2/(aperture_diameter_x/2.35)**2) + \
                                                  ((Y*Y)/2/(aperture_diameter_y/2.35)**2) )))

        return wavefront

class WOSlit1D(Slit, OpticalElementDecorator):
    def __init__(self, name="Undefined", boundary_shape=BoundaryShape()):
        Slit.__init__(self, name=name, boundary_shape=boundary_shape)

    def applyOpticalElement(self, wavefront, parameters=None, element_index=None):
        boundaries = self._boundary_shape.get_boundaries()

        if isinstance(self._boundary_shape, Rectangle):
            wavefront.clip(boundaries[0], boundaries[1])
        else:
            raise NotImplementedError("to be implemented")

        return wavefront

class WOGaussianSlit1D(Slit, OpticalElementDecorator):
    def __init__(self, name="Undefined", boundary_shape=BoundaryShape()):
        Slit.__init__(self, name=name, boundary_shape=boundary_shape)

    def applyOpticalElement(self, wavefront, parameters=None, element_index=None):
        boundaries = self._boundary_shape.get_boundaries()
        aperture_diameter =  numpy.abs(boundaries[1] - boundaries[0])
        X = wavefront.get_abscissas()

        window = numpy.exp(-(X*X)/2/(aperture_diameter/2.35)**2)
        wavefront.rescale_amplitudes(window)

        return wavefront



#   propagate_2D_fresnel               \
#   propagate_2D_fresnel_convolution   | Near field Fresnel propagators via convolution in Fourier space. Three methods
#   propagate_2D_fresnel_srw          /
#
#          three methods available: 'fft': fft -> multiply by kernel in freq -> ifft
#                                   'convolution': scipy.signal.fftconvolve(wave,kernel in space)
#                                   'srw': use the SRW package
#
#
#
# *********************************** IMPORTANT *******************************************
#                RECOMMENDATIONS:
#
#     >>> Prefer propagate_2D_fresnel <<<
#       Prefer EVEN number of bins.
#       Set shift_half_pixel=1 (now the default)
#    Under these circumstances, the results agree very well with SRW
#
#
#


import numpy

from wofry.propagator.wavefront2D.generic_wavefront import GenericWavefront2D
from wofry.propagator.propagator import Propagator2D

class FresnelConvolution2D(Propagator2D):

    HANDLER_NAME = "FRESNEL_CONVOLUTION_2D"

    def get_handler_name(self):
        return self.HANDLER_NAME


    def do_specific_progation_after(self, wavefront, propagation_distance, parameters, element_index=None):
        return self.do_specific_progation(wavefront, propagation_distance, parameters, element_index=element_index)

    def do_specific_progation_before(self, wavefront, propagation_distance, parameters, element_index=None):
        return self.do_specific_progation( wavefront, propagation_distance, parameters, element_index=element_index)

    """
    2D Fresnel propagator using convolution via Fourier transform
    :param wavefront:
    :param propagation_distance: propagation distance
    :param shift_half_pixel: set to 1 to shift half pixel (recommended using an even number of pixels) Set as default.
    :return: a new 2D wavefront object with propagated wavefront
    """

    def do_specific_progation(self, wavefront, propagation_distance, parameters, element_index=None):

        shift_half_pixel = self.get_additional_parameter("shift_half_pixel",False,parameters,element_index=element_index)

        is_generic_wavefront = isinstance(wavefront, GenericWavefront2D)

        if is_generic_wavefront:
            pass
        else:
            wavefront = wavefront.toGenericWavefront()

        return self.propagate_wavefront(wavefront,propagation_distance,shift_half_pixel=shift_half_pixel)

    @classmethod
    def propagate_wavefront(cls,wavefront,propagation_distance,shift_half_pixel=False):

        from scipy.signal import fftconvolve

        wavelength = wavefront.get_wavelength()

        X = wavefront.get_mesh_x()
        Y = wavefront.get_mesh_y()

        if shift_half_pixel:
            x = wavefront.get_coordinate_x()
            y = wavefront.get_coordinate_y()
            X += 0.5 * numpy.abs( x[0] - x[1] )
            Y += 0.5 * numpy.abs( y[0] - y[1] )

        kernel = numpy.exp(1j*2*numpy.pi/wavefront.get_wavelength() *
                           (X**2 + Y**2) / 2 / propagation_distance)
        kernel *= numpy.exp(1j*2*numpy.pi/wavefront.get_wavelength() * propagation_distance)
        kernel /=  1j * wavefront.get_wavelength() * propagation_distance

        wavefront_out = GenericWavefront2D.initialize_wavefront_from_arrays(x_array=wavefront.get_coordinate_x(),
                                                                            y_array=wavefront.get_coordinate_y(),
                                                                            z_array=fftconvolve(wavefront.get_complex_amplitude(),
                                                                                                kernel,
                                                                                                mode='same'),
                                                                            wavelength=wavelength)
        # added srio@esrf.eu 2018-03-23 to conserve energy - TODO: review method!
        wavefront_out.rescale_amplitude( numpy.sqrt(wavefront.get_intensity().sum() /
                                                    wavefront_out.get_intensity().sum()))

        return wavefront_out



###################################################################
# DO NOT TOUCH THIS CODE -- BEGIN
###################################################################
import threading

def synchronized_method(method):

    outer_lock = threading.Lock()
    lock_name = "__"+method.__name__+"_lock"+"__"

    def sync_method(self, *args, **kws):
        with outer_lock:
            if not hasattr(self, lock_name): setattr(self, lock_name, threading.Lock())
            lock = getattr(self, lock_name)
            with lock:
                return method(self, *args, **kws)

    return sync_method

class Singleton:

    def __init__(self, decorated):
        self._decorated = decorated

    @synchronized_method
    def Instance(self):
        try:
            return self._instance
        except AttributeError:
            self._instance = self._decorated()
            return self._instance

    def __call__(self):
        raise TypeError('Singletons must be accessed through `Instance()`.')

    def __instancecheck__(self, inst):
        return isinstance(inst, self._decorated)

###################################################################
# DO NOT TOUCH THIS CODE -- END
###################################################################


from syned.beamline.beamline_element import BeamlineElement

from wofry.propagator.wavefront  import Wavefront, WavefrontDimension

class PropagationElements(object):
    
    INSERT_AFTER = 0
    INSERT_BEFORE = 1

    def __init__(self):
        self.__propagation_elements = []
        self.__propagation_element_parameters = []

    def add_beamline_element(self, beamline_element=BeamlineElement(), element_parameters=None):
        if beamline_element is None: raise ValueError("Beamline is None")

        self.__propagation_elements.append(beamline_element)
        self.__propagation_element_parameters.append(element_parameters)

    def add_beamline_elements(self, beamline_elements=[], element_parameters_list=None):
        if beamline_elements is None: raise ValueError("Beamline is None")
        
        if not element_parameters_list is None:
            if len(beamline_elements) != len(element_parameters_list): raise ValueError("Specific Parameters list does not match Beamline Elements list")
        else:
            element_parameters_list = [None]*len(beamline_elements)
            
        for beamline_element, element_parameters in zip(beamline_elements, element_parameters_list):
            self.add_beamline_element(beamline_element, element_parameters)
        
    def insert_beamline_element(self, index, new_element=BeamlineElement(), mode=INSERT_BEFORE, new_element_parameters=None):
        if mode == PropagationElements.INSERT_BEFORE:
            if index == 0:
                self.__propagation_elements = [new_element] + self.__propagation_elements
                self.__propagation_element_parameters = [new_element_parameters]
            else:
                self.__propagation_elements.insert(index, new_element)
                self.__propagation_element_parameters.insert(index, new_element_parameters)
        elif mode == PropagationElements.INSERT_AFTER:
            if index == len(self.__propagation_elements) - 1:
                self.__propagation_elements = self.__propagation_elements + [new_element]
                self.__propagation_element_parameters = self.__propagation_element_parameters + [new_element_parameters]
            else:
                self.__propagation_elements.insert(index+1, new_element)
                self.__propagation_element_parameters.insert(index+1, new_element_parameters)

    def get_propagation_elements_number(self):
        return len(self.__propagation_elements)

    def get_propagation_elements(self):
        return self.__propagation_elements

    def get_propagation_element(self, index):
        return self.__propagation_elements[index]

    def get_propagation_elements_parameters(self):
        return self.__propagation_element_parameters

    def get_propagation_element_parameter(self, index):
        return self.__propagation_element_parameters[index]



class PropagationParameters(object):
    def __init__(self,
                 wavefront = Wavefront(),
                 propagation_elements = PropagationElements()):
        self._wavefront = wavefront
        self._propagation_elements = propagation_elements
        self._additional_parameters = None

    def get_wavefront(self):
        return self._wavefront

    def get_PropagationElements(self):
        return self._propagation_elements

    def set_additional_parameters(self, key, value):
        if self._additional_parameters is None:
            self._additional_parameters = {key : value}
        else:
            self._additional_parameters[key] = value

    def get_additional_parameter(self, key):
        return self._additional_parameters[key]

    def has_additional_parameter(self, key):
        return key in self._additional_parameters

class AbstractPropagator(object):

    def __init__(self):
        super().__init__()

    def get_dimension(self):
        raise NotImplementedError("This method is abstract")

    def get_handler_name(self):
        raise NotImplementedError("This method is abstract")

    def is_handler(self, handler_name):
        return handler_name == self.get_handler_name()

    def do_propagation(self, parameters=PropagationParameters()):
        raise NotImplementedError("This method is abstract" +
                                  "\n\naccepts " + PropagationParameters.__module__ + "." + PropagationParameters.__name__ +
                                  "\nreturns " + Wavefront.__module__ + "." + Wavefront.__name__)


class PropagationMode:
    STEP_BY_STEP = 0
    WHOLE_BEAMLINE = 1

class PropagationApplication:
    ALL = "All"

@Singleton
class PropagationManager(object):

    def __init__(self):
        self.__chains_hashmap = {WavefrontDimension.ONE : [],
                                 WavefrontDimension.TWO : []}

        self.__propagation_mode_hashmap = {PropagationApplication.ALL : PropagationMode.STEP_BY_STEP}
        self.__is_initialized_hashmap = {PropagationApplication.ALL : False}

    @synchronized_method
    def set_initialized(self, application = PropagationApplication.ALL, initialized=True):
        self.__is_initialized_hashmap[application] = initialized

    @synchronized_method
    def is_initialized(self, application = PropagationApplication.ALL):
        if application in self.__is_initialized_hashmap.keys():
            return self.__is_initialized_hashmap[application]
        else:
            return False

    @synchronized_method
    def set_propagation_mode(self, application = PropagationApplication.ALL, mode=PropagationMode.STEP_BY_STEP):
        self.__propagation_mode_hashmap[application] = mode

    @synchronized_method
    def get_propagation_mode(self, application = PropagationApplication.ALL):
        return self.__propagation_mode_hashmap[application]

    @synchronized_method
    def add_propagator(self, propagator=AbstractPropagator()):
        if propagator is None: raise ValueError("Given propagator is None")
        if not isinstance(propagator, AbstractPropagator): raise ValueError("Given propagator is not a compatible object")

        dimension = propagator.get_dimension()

        if not (dimension == WavefrontDimension.ONE or dimension == WavefrontDimension.TWO):
            raise ValueError("Wrong propagator dimension")

        propagation_chain_of_responsibility = self.__chains_hashmap.get(dimension)

        for existing in propagation_chain_of_responsibility:
            if existing.is_handler(propagator.get_handler_name()):
                raise ValueError("Propagator already in the Chain")

        propagation_chain_of_responsibility.append(propagator)

    @synchronized_method
    def has_propagator(self, handler_name="<Propagator Name>", dimension=WavefrontDimension.ONE):
        propagation_chain_of_responsibility = self.__chains_hashmap.get(dimension)

        for existing in propagation_chain_of_responsibility:
            if existing.get_handler_name() == handler_name: return True

        return False

    @synchronized_method
    def get_propagators_number(self, dimension=None):
        propagation_chain_of_responsibility_1D = self.__chains_hashmap.get(WavefrontDimension.ONE)
        propagation_chain_of_responsibility_2D = self.__chains_hashmap.get(WavefrontDimension.TWO)

        if dimension == None:
            return len(propagation_chain_of_responsibility_1D), len(propagation_chain_of_responsibility_2D)
        elif dimension == WavefrontDimension.ONE:
            return len(propagation_chain_of_responsibility_1D)
        elif dimension == WavefrontDimension.TWO:
            return len(propagation_chain_of_responsibility_2D)
        else:
            raise ValueError("Dimension not valid " + str(dimension))

    def do_propagation(self, propagation_parameters, handler_name):
        for propagator in self.__chains_hashmap.get(propagation_parameters.get_wavefront().get_dimension()):
            if propagator.is_handler(handler_name):
                return propagator.do_propagation(parameters=propagation_parameters)

        raise Exception("Handler not found: "+handler_name)

# ---------------------------------------------------------------

class Propagator(AbstractPropagator):

    def do_propagation(self, parameters=PropagationParameters()):
        wavefront = parameters.get_wavefront()

        for index in range(0, parameters.get_PropagationElements().get_propagation_elements_number()):
            element = parameters.get_PropagationElements().get_propagation_element(index)
            coordinates = element.get_coordinates()

            if coordinates.p() != 0.0: wavefront = self.do_specific_progation_before(wavefront, coordinates.p(), parameters, element_index=index)
            wavefront = element.get_optical_element().applyOpticalElement(wavefront, parameters, element_index=index)
            if coordinates.q() != 0.0: wavefront = self.do_specific_progation_after(wavefront, coordinates.q(), parameters, element_index=index)


        return wavefront

    def do_specific_progation_before(self, wavefront, propagation_distance, parameters, element_index=None):
        raise NotImplementedError("This method is abstract")

    def do_specific_progation_after(self, wavefront, propagation_distance, parameters, element_index=None):
        raise NotImplementedError("This method is abstract")

    def get_additional_parameter(self, parameter_name, default_value, propagation_parameters, element_index=None, ):
        value = default_value
        try:
            value = propagation_parameters.get_additional_parameter(parameter_name)
        except:
            pass

        if element_index is None:
            myindex = 0
        else:
            myindex = element_index

        parameters_dict = propagation_parameters.get_PropagationElements().get_propagation_element_parameter(myindex)

        if parameters_dict is not None:
            if parameter_name in parameters_dict.keys():
                value = parameters_dict[parameter_name]

        return value


class Propagator1D(Propagator):

    def get_dimension(self):
        return WavefrontDimension.ONE

    def do_propagation(self, parameters=PropagationParameters()):
        if not parameters.get_wavefront().get_dimension() == WavefrontDimension.ONE:
            raise Exception("wrong wavefront!  it is not 1D")

        return super().do_propagation(parameters)

class Propagator2D(Propagator):

    def get_dimension(self):
        return WavefrontDimension.TWO

    def do_propagation(self, parameters=PropagationParameters()):
        if not parameters.get_wavefront().get_dimension() == WavefrontDimension.TWO:
            raise Exception("wrong wavefront!  it is not 2D")

        return super().do_propagation(parameters)

